@@ -65,6 +65,9 @@ The type of this setting:
  - $CONF_TYPE_HASH_KEY_VALUE: hash key/value pair, like "describe" or tflags
  - $CONF_TYPE_STRINGLIST list of strings, stored as an array
  - $CONF_TYPE_IPADDRLIST list of IP addresses, stored as an array of SA::NetSet
+ - $CONF_TYPE_DURATION a nonnegative time interval in seconds - a numeric value
+                      (float or int), optionally suffixed by a time unit (s, m,
+                      h, d, w), seconds are implied if unit is missing
 
 If this is set, and a 'code' block does not already exist, a 'code' block is
 assigned based on the type.
@@ -77,10 +80,10 @@ value be non-empty, otherwise they'll produce a warning message.
 
 =item code
 
-A subroutine to deal with the setting.  Only used if B<type> is not set.  ONE OF
-B<code> OR B<type> IS REQUIRED.  The arguments passed to the function are
-C<($self, $key, $value, $line)>, where $key is the setting (*not* the command),
-$value is the value string, and $line is the entire line.
+A subroutine to deal with the setting.  ONE OF B<code> OR B<type> IS REQUIRED.
+The arguments passed to the function are C<($self, $key, $value, $line)>,
+where $key is the setting (*not* the command), $value is the value string,
+and $line is the entire line.
 
 There are two special return values that the B<code> subroutine may return
 to signal that there is an error in the configuration:
@@ -91,6 +94,10 @@ that a value be set, but one was not provided.
 C<$Mail::SpamAssassin::Conf::INVALID_VALUE> -- this setting requires a value
 from a set of 'valid' values, but the user provided an invalid one.
 
+C<$Mail::SpamAssassin::Conf::INVALID_HEADER_FIELD_NAME> -- this setting
+requires a syntactically valid header field name, but the user provided
+an invalid one.
+
 Any other values -- including C<undef> -- returned from the subroutine are
 considered to mean 'success'.
 
@@ -130,19 +137,18 @@ package Mail::SpamAssassin::Conf::Parser;
 use Mail::SpamAssassin::Conf;
 use Mail::SpamAssassin::Constants qw(:sa);
 use Mail::SpamAssassin::Logger;
-use Mail::SpamAssassin::Util qw(untaint_var);
+use Mail::SpamAssassin::Util qw(untaint_var compile_regexp);
 use Mail::SpamAssassin::NetSet;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 
-use vars qw{
-  @ISA
-};
+our @ISA = qw();
 
-@ISA = qw();
+my $ARITH_EXPRESSION_LEXER = ARITH_EXPRESSION_LEXER;
+my $META_RULES_MATCHING_RE = META_RULES_MATCHING_RE;
 
 ###########################################################################
 
@@ -215,8 +221,8 @@ sub build_command_luts {
 sub parse {
   my ($self, undef, $scoresonly) = @_; # leave $rules in $_[1]
 
-  $self->{scoresonly} = $scoresonly;
   my $conf = $self->{conf};
+  $self->{scoresonly} = $scoresonly;
 
   # Language selection:
   # See http://www.gnu.org/manual/glibc-2.2.5/html_node/Locale-Categories.html
@@ -256,14 +262,24 @@ sub parse {
   while (defined ($line = shift @conf_lines)) {
     local ($1);         # bug 3838: prevent random taint flagging of $1
 
-    # bug 5545: used to support testing rules in the ruleqa system
-    if ($keepmetadata && $line =~ /^\#testrules/) {
-      $self->{file_scoped_attrs}->{testrules}++;
-      next;
+    if (index($line,'#') > -1) {
+      # bug 5545: used to support testing rules in the ruleqa system
+      if ($keepmetadata && $line =~ /^\#testrules/) {
+        $self->{file_scoped_attrs}->{testrules}++;
+        next;
+      }
+
+      # bug 6800: let X-Spam-Checker-Version also show what sa-update we are at
+      if ($line =~ /^\# UPDATE version (\d+)$/) {
+        for ($self->{currentfile}) {  # just aliasing, not a loop
+          $conf->{update_version}{$_} = $1  if defined $_ && $_ ne '(no file)';
+        }
+      }
+
+      $line =~ s/(?<!\\)#.*$//; # remove comments
+      $line =~ s/\\#/#/g; # hash chars are escaped, so unescape them
     }
 
-    $line =~ s/(?<!\\)#.*$//; # remove comments
-    $line =~ s/\\#/#/g; # hash chars are escaped, so unescape them
     $line =~ s/^\s+//;  # remove leading whitespace
     $line =~ s/\s+$//;  # remove tailing whitespace
     next unless($line); # skip empty lines
@@ -274,7 +290,7 @@ sub parse {
     my($key, $value) = split(/\s+/, $line, 2);
     $key = lc $key;
     # convert all dashes in setting name to underscores.
-    $key =~ s/-/_/g;
+    $key =~ tr/-/_/;
     $value = '' unless defined($value);
 
 #   # Do a better job untainting this info ...
@@ -324,26 +340,26 @@ sub parse {
     }
 
     # now handle the commands.
-    if ($key eq 'include') {
+    elsif ($key eq 'include') {
       $value = $self->fix_path_relative_to_current_file($value);
       my $text = $conf->{main}->read_cf($value, 'included file');
       unshift (@conf_lines, split (/\n/, $text));
       next;
     }
 
-    if ($key eq 'ifplugin') {
+    elsif ($key eq 'ifplugin') {
       $self->handle_conditional ($key, "plugin ($value)",
                         \@if_stack, \$skip_parsing);
       next;
     }
 
-    if ($key eq 'if') {
+    elsif ($key eq 'if') {
       $self->handle_conditional ($key, $value,
                         \@if_stack, \$skip_parsing);
       next;
     }
 
-    if ($key eq 'else') {
+    elsif ($key eq 'else') {
       # TODO: if/else/else won't get flagged here :(
       if (!@if_stack) {
         $parse_error = "config: found else without matching conditional";
@@ -355,7 +371,7 @@ sub parse {
     }
 
     # and the endif statement:
-    if ($key eq 'endif') {
+    elsif ($key eq 'endif') {
       my $lastcond = pop @if_stack;
       if (!defined $lastcond) {
         $parse_error = "config: found endif without matching conditional";
@@ -379,7 +395,7 @@ sub parse {
       # "3.0.x" versions:
       ## make sure it's a numeric value
       #$value += 0.0;
-      ## convert 3.000000 -> 3.0, stay backwards compatible ...
+      ## convert 3.000000 -> 3.0, stay backward compatible ...
       #$ver =~ s/^(\d+)\.(\d{1,3}).*$/sprintf "%d.%d", $1, $2/e;
       #$value =~ s/^(\d+)\.(\d{1,3}).*$/sprintf "%d.%d", $1, $2/e;
 
@@ -430,6 +446,13 @@ sub parse {
                         "skipping: $line";
         goto failed_line;
       }
+      elsif ($ret && $ret eq $Mail::SpamAssassin::Conf::INVALID_HEADER_FIELD_NAME)
+      {
+        $parse_error = "config: SpamAssassin failed to parse line, ".
+                       "it does not specify a valid header field name, ".
+                       "skipping: $line";
+        goto failed_line;
+      }
       elsif ($ret && $ret eq $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE)
       {
         $parse_error = "config: SpamAssassin failed to parse line, ".
@@ -477,6 +500,7 @@ failed_line:
 
   $self->lint_check();
   $self->set_default_scores();
+  $self->check_for_missing_descriptions();
 
   delete $self->{scoresonly};
 }
@@ -485,37 +509,54 @@ sub handle_conditional {
   my ($self, $key, $value, $if_stack_ref, $skip_parsing_ref) = @_;
   my $conf = $self->{conf};
 
-  my $lexer = ARITH_EXPRESSION_LEXER;
-  my @tokens = ($value =~ m/($lexer)/g);
+  my @tokens = ($value =~ /($ARITH_EXPRESSION_LEXER)/og);
 
   my $eval = '';
   my $bad = 0;
   foreach my $token (@tokens) {
-    if ($token =~ /^(?:\W+|[+-]?\d+(?:\.\d+)?)$/) {
-      $eval .= untaint_var($token) . " ";
+    if ($token =~ /^(?:\W{1,5}|[+-]?\d+(?:\.\d+)?)$/) {
+      # using tainted subr. argument may taint the whole expression, avoid
+      my $u = untaint_var($token);
+      $eval .= $u . " ";
     }
     elsif ($token eq 'plugin') {
-      # replace with method call
-      $eval .= "\$self->cond_clause_plugin_loaded";
+      # replace with a method call
+      $eval .= '$self->cond_clause_plugin_loaded';
     }
     elsif ($token eq 'can') {
-      # replace with method call
-      $eval .= "\$self->cond_clause_can";
+      # replace with a method call
+      $eval .= '$self->cond_clause_can';
+    }
+    elsif ($token eq 'has') {
+      # replace with a method call
+      $eval .= '$self->cond_clause_has';
     }
     elsif ($token eq 'version') {
       $eval .= $Mail::SpamAssassin::VERSION." ";
     }
+    elsif ($token eq 'perl_version') {
+      $eval .= $]." ";
+    }
     elsif ($token =~ /^\w[\w\:]+$/) { # class name
-      $eval .= '"' . untaint_var($token) . '" ';
+      # Strictly controlled form:
+      if ($token =~ /^(?:\w+::){0,10}\w+$/) {
+        my $u = untaint_var($token);
+        $eval .= "'$u'";
+      } else {
+        warn "config: illegal name '$token' in 'if $value'\n";
+        $bad++;
+        last;
+      }
     }
     else {
       $bad++;
       warn "config: unparseable chars in 'if $value': '$token'\n";
+      last;
     }
   }
 
   if ($bad) {
-    $self->lint_warn("bad 'if' line, in \"$self->{currentfile}\"", undef);
+    $self->lint_warn("config: bad 'if' line, in \"$self->{currentfile}\"", undef);
     return -1;
   }
 
@@ -541,15 +582,35 @@ sub cond_clause_plugin_loaded {
 
 sub cond_clause_can {
   my ($self, $method) = @_;
+  if ($self->{currentfile} =~ q!\buser_prefs$! ) {
+    warn "config: 'if can $method' not available in user_prefs";
+    return 0
+  }
+  $self->cond_clause_can_or_has('can', $method);
+}
+
+sub cond_clause_has {
+  my ($self, $method) = @_;
+  $self->cond_clause_can_or_has('has', $method);
+}
+
+sub cond_clause_can_or_has {
+  my ($self, $fn_name, $method) = @_;
 
   local($1,$2);
-  if ($method =~ /^(.*)::([^:]+)$/) {
+  if (!defined $method) {
+    $self->lint_warn("config: bad 'if' line, no argument to $fn_name(), ".
+                     "in \"$self->{currentfile}\"", undef);
+  } elsif ($method =~ /^(.*)::([^:]+)$/) {
+    no strict "refs";
     my($module, $meth) = ($1, $2);
-    return UNIVERSAL::can($module, $meth);
+    return 1  if $module->can($meth) &&
+                 ( $fn_name eq 'has' || &{$method}() );
   } else {
-    $self->lint_warn("bad 'if' line, cannot find '::' in can($method), ".
-                "in \"$self->{currentfile}\"", undef);
+    $self->lint_warn("config: bad 'if' line, cannot find '::' in $fn_name($method), ".
+                     "in \"$self->{currentfile}\"", undef);
   }
+  return;
 }
 
 # Let's do some linting here ...
@@ -564,7 +625,7 @@ sub lint_check {
     # Check for description and score issues in lint fashion
     while ( my $k = each %{$conf->{descriptions}} ) {
       if (!exists $conf->{tests}->{$k}) {
-        $self->lint_warn("config: warning: description exists for non-existent rule $k\n", $k);
+        dbg("config: warning: description exists for non-existent rule $k");
       }
     }
 
@@ -599,6 +660,21 @@ sub set_default_scores {
   }
 }
 
+# loop through all the tests and if we are missing a description with debug
+# set, throw a warning except for testing T_ or meta __ rules.
+sub check_for_missing_descriptions {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+
+  while ( my $k = each %{$conf->{tests}} ) {
+    if ($k !~ m/^(?:T_|__)/i) {
+      if ( ! exists $conf->{descriptions}->{$k} ) {
+        dbg("config: warning: no description set for $k");
+      }
+    }
+  }
+}
+
 ###########################################################################
 
 sub setup_default_code_cb {
@@ -632,6 +708,9 @@ sub setup_default_code_cb {
   elsif ($type == $Mail::SpamAssassin::Conf::CONF_TYPE_IPADDRLIST) {
     $cmd->{code} = \&set_ipaddr_list;
   }
+  elsif ($type == $Mail::SpamAssassin::Conf::CONF_TYPE_DURATION) {
+    $cmd->{code} = \&set_duration_value;
+  }
   else {
     warn "config: unknown conf type $type!";
     return 0;
@@ -653,11 +732,28 @@ sub set_numeric_value {
   unless (defined $value && $value !~ /^$/) {
     return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
   }
-  unless ($value =~ /^-?\d+(?:\.\d+)?$/) {
+  unless ($value =~ /^ [+-]? \d+ (?: \. \d* )? \z/sx) {
     return $Mail::SpamAssassin::Conf::INVALID_VALUE;
   }
-  # it is safe to untaint now that we now the syntax is a valid number
-  $conf->{$key} = untaint_var($value) + 0.0;
+  # it is safe to untaint now that we know the syntax is a valid number
+  $conf->{$key} = untaint_var($value) + 0;
+}
+
+sub set_duration_value {
+  my ($conf, $key, $value, $line) = @_;
+
+  local ($1,$2);
+  unless (defined $value && $value !~ /^$/) {
+    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+  }
+  unless ($value =~ /^( \+? \d+ (?: \. \d* )? ) (?: \s* ([smhdw]))? \z/sxi) {
+    return $Mail::SpamAssassin::Conf::INVALID_VALUE;
+  }
+  $value = $1;
+  $value *= { s => 1, m => 60, h => 3600,
+              d => 24*3600, w => 7*24*3600 }->{lc $2}  if defined $2;
+  # it is safe to untaint now that we know the syntax is a valid time interval
+  $conf->{$key} = untaint_var($value) + 0;
 }
 
 sub set_bool_value {
@@ -790,39 +886,40 @@ sub finish_parsing {
 
     # eval type handling
     if (($type & 1) == 1) {
-      if (my ($function, $args) = ($text =~ m/(.*?)\s*\((.*?)\)\s*$/)) {
-        my ($packed, $argsref) =
-                $self->pack_eval_method($function, $args, $name, $text);
-
-        if (!$packed) {
-          # we've already warned about this
+      if (my ($function, $args) = ($text =~ /^(\w+)\((.*?)\)$/)) {
+        my $argsref = $self->pack_eval_args($args);
+        if (!defined $argsref) {
+          $self->lint_warn("syntax error for eval function $name: $text");
+          next;
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS) {
-          $conf->{body_evals}->{$priority}->{$name} = $packed;
+          $conf->{body_evals}->{$priority}->{$name} = [ $function, [@$argsref] ];
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS) {
-          $conf->{head_evals}->{$priority}->{$name} = $packed;
+          $conf->{head_evals}->{$priority}->{$name} = [ $function, [@$argsref] ];
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_RBL_EVALS) {
           # We don't do priorities for $Mail::SpamAssassin::Conf::TYPE_RBL_EVALS
           # we also use the arrayref instead of the packed string
-          $conf->{rbl_evals}->{$name} = [ $function, @$argsref ];
+          $conf->{rbl_evals}->{$name} = [ $function, [@$argsref] ];
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_RAWBODY_EVALS) {
-          $conf->{rawbody_evals}->{$priority}->{$name} = $packed;
+          $conf->{rawbody_evals}->{$priority}->{$name} = [ $function, [@$argsref] ];
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_FULL_EVALS) {
-          $conf->{full_evals}->{$priority}->{$name} = $packed;
+          $conf->{full_evals}->{$priority}->{$name} = [ $function, [@$argsref] ];
         }
         #elsif ($type == $Mail::SpamAssassin::Conf::TYPE_URI_EVALS) {
-        #  $conf->{uri_evals}->{$priority}->{$name} = $packed;
+        #  $conf->{uri_evals}->{$priority}->{$name} = [ $function, [@$argsref] ];
         #}
         else {
           $self->lint_warn("unknown type $type for $name: $text", $name);
+          next;
         }
       }
       else {
         $self->lint_warn("syntax error for eval function $name: $text", $name);
+        next;
       }
     }
     # non-eval tests
@@ -849,6 +946,7 @@ sub finish_parsing {
       }
       else {
         $self->lint_warn("unknown type $type for $name: $text", $name);
+        next;
       }
     }
   }
@@ -867,7 +965,8 @@ sub finish_parsing {
     # free up stuff we no longer need
     delete $conf->{tests};
     delete $conf->{priority};
-    delete $conf->{test_types};
+    #test_types are needed - see bug 5503
+    #delete $conf->{test_types};
   }
 }
 
@@ -879,40 +978,49 @@ sub trace_meta_dependencies {
   foreach my $name (keys %{$conf->{tests}}) {
     next unless ($conf->{test_types}->{$name}
                     == $Mail::SpamAssassin::Conf::TYPE_META_TESTS);
-
-    my $deps = [ ];
-    my $alreadydone = { };
-    $self->_meta_deps_recurse($conf, $name, $name, $deps, $alreadydone);
-    $conf->{meta_dependencies}->{$name} = join (' ', @{$deps});
+    my $alreadydone = {};
+    $self->_meta_deps_recurse($conf, $name, $name, $alreadydone);
   }
 }
 
 sub _meta_deps_recurse {
-  my ($self, $conf, $toprule, $name, $deps, $alreadydone) = @_;
+  my ($self, $conf, $toprule, $name, $alreadydone) = @_;
 
-  # Only do each rule once per top-level meta; avoid infinite recursion
-  return if $alreadydone->{$name};
-  $alreadydone->{$name} = 1;
+  # Avoid recomputing the dependencies of a rule
+  return split(' ', $conf->{meta_dependencies}->{$name}) if defined $conf->{meta_dependencies}->{$name};
 
   # Obviously, don't trace empty or nonexistent rules
   my $rule = $conf->{tests}->{$name};
-  return unless $rule;
+  unless ($rule) {
+      $conf->{meta_dependencies}->{$name} = '';
+      return ( );
+  }
+
+  # Avoid infinite recursion
+  return ( ) if exists $alreadydone->{$name};
+  $alreadydone->{$name} = ( );
+
+  my %deps;
 
   # Lex the rule into tokens using a rather simple RE method ...
-  my $lexer = ARITH_EXPRESSION_LEXER;
-  my @tokens = ($rule =~ m/$lexer/g);
+  my @tokens = ($rule =~ /($ARITH_EXPRESSION_LEXER)/og);
 
   # Go through each token in the meta rule
+  my $conf_tests = $conf->{tests};
   foreach my $token (@tokens) {
     # has to be an alpha+numeric token
-    next if ($token =~ /^(?:\W+|[+-]?\d+(?:\.\d+)?)$/);
+    next if $token =~ tr{A-Za-z0-9_}{}c || substr($token,0,1) =~ tr{A-Za-z_}{}c; # even faster
+
     # and has to be a rule name
-    next unless exists $conf->{tests}->{$token};
+    next unless exists $conf_tests->{$token};
 
     # add and recurse
-    push(@{$deps}, untaint_var($token));
-    $self->_meta_deps_recurse($conf, $toprule, $token, $deps, $alreadydone);
+    $deps{untaint_var($token)} = ( );
+    my @subdeps = $self->_meta_deps_recurse($conf, $toprule, $token, $alreadydone);
+    @deps{@subdeps} = ( );
   }
+  $conf->{meta_dependencies}->{$name} = join (' ', keys %deps);
+  return keys %deps;
 }
 
 sub fix_priorities {
@@ -997,40 +1105,36 @@ sub find_dup_rules {
   }
 }
 
+# Deprecated function
 sub pack_eval_method {
-  my ($self, $function, $args, $name, $text) = @_;
+  warn "deprecated function pack_eval_method() used\n";
+  return ('',undef);
+}
 
+sub pack_eval_args {
+  my ($self, $args) = @_;
+
+  return [] if $args =~ /^\s+$/;
+
+  # bug 4419: Parse quoted strings, unquoted alphanumerics/floats,
+  # unquoted IPv4 and IPv6 addresses, and unquoted common domain names.
+  # s// is used so that we can determine whether or not we successfully
+  # parsed ALL arguments.
   my @args;
-  if (defined $args) {
-    # bug 4419: Parse quoted strings, unquoted alphanumerics/floats,
-    # unquoted IPv4 and IPv6 addresses, and unquoted common domain names.
-    # s// is used so that we can determine whether or not we successfully
-    # parsed ALL arguments.
-    local($1,$2,$3);
-    while ($args =~ s/^\s* (?: (['"]) (.*?) \1 | ( [\d\.:A-Za-z-]+? ) )
-                       \s* (?: , \s* | $ )//x) {
-      if (defined $2) {
-        push @args, $2;
-      }
-      else {
-        push @args, $3;
-      }
-    }
+  local($1,$2,$3);
+  while ($args =~ s/^\s* (?: (['"]) (.*?) \1 | ( [\d\.:A-Za-z-]+? ) )
+                     \s* (?: , \s* | $ )//x) {
+    # DO NOT UNTAINT THESE ARGS
+    # The eval function that handles these should do that as necessary,
+    # we have no idea what acceptable arguments look like here.
+    push @args, defined $2 ? $2 : $3;
   }
 
   if ($args ne '') {
-    $self->lint_warn("syntax error (unparsable argument: $args) for eval function: $name: $text", $name);
-    return;
+    return undef; ## no critic (ProhibitExplicitReturnUndef)
   }
 
-  my $argstr = $function;
-  $argstr =~ s/\s+//gs;
-
-  if (@args > 0) {
-    $argstr .= ',' . join(', ',
-              map { my $s = $_; $s =~ s/\#/[HASH]/gs; 'q#' . $s . '#' } @args);
-  }
-  return ($argstr, \@args);
+  return \@args;
 }
 
 ###########################################################################
@@ -1092,55 +1196,101 @@ sub add_test {
   my $conf = $self->{conf};
 
   # Don't allow invalid names ...
-  if ($name !~ /^\D\w*$/) {
+  if ($name !~ IS_RULENAME) {
     $self->lint_warn("config: error: rule '$name' has invalid characters ".
 	   "(not Alphanumeric + Underscore + starting with a non-digit)\n", $name);
     return;
   }
 
-  # Also set a hard limit for ALL rules (rule names longer than 242
+  # Also set a hard limit for ALL rules (rule names longer than 40
   # characters throw warnings).  Check this separately from the above
   # pattern to avoid vague error messages.
-  if (length $name > 200) {
-    $self->lint_warn("config: error: rule '$name' is way too long ".
+  if (length $name > 100) {
+    $self->lint_warn("config: error: rule '$name' is too long ".
 	   "(recommended maximum length is 22 characters)\n", $name);
     return;
   }
 
   # Warn about, but use, long rule names during --lint
   if ($conf->{lint_rules}) {
-    if (length($name) > 50 && $name !~ /^__/ && $name !~ /^T_/) {
-      $self->lint_warn("config: warning: rule name '$name' is over 50 chars ".
+    if (length($name) > 40 && $name !~ /^__/ && $name !~ /^T_/) {
+      $self->lint_warn("config: warning: rule name '$name' is over 40 chars ".
 	     "(recommended maximum length is 22 characters)\n", $name);
     }
   }
 
+  # parameter to compile_regexp()
+  my $ignore_amre =
+    $self->{conf}->{lint_rules} ||
+    $self->{conf}->{ignore_always_matching_regexps};
+
   # all of these rule types are regexps
   if ($type == $Mail::SpamAssassin::Conf::TYPE_BODY_TESTS ||
       $type == $Mail::SpamAssassin::Conf::TYPE_FULL_TESTS ||
       $type == $Mail::SpamAssassin::Conf::TYPE_RAWBODY_TESTS ||
       $type == $Mail::SpamAssassin::Conf::TYPE_URI_TESTS)
   {
-    return unless $self->is_delimited_regexp_valid($name, $text);
+    my ($rec, $err) = compile_regexp($text, 1, $ignore_amre);
+    if (!$rec) {
+      $self->lint_warn("config: invalid regexp for $name '$text': $err", $name);
+      return;
+    }
+    $conf->{test_qrs}->{$name} = $rec;
   }
-  if ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_TESTS)
+  elsif ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_TESTS)
   {
+    local($1,$2,$3);
     # RFC 5322 section 3.6.8, ftext printable US-ASCII chars not including ":"
-    if ($text =~ /^!?defined\([!-9;-\176]+\)$/) {
-      # fine, implements 'exists:'
+    # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
+    if ($text =~ /^exists:(.*)/) {
+      my $hdr = $1;
+      # never evaled, so can be quite generous with the name
+      # check :addr etc header options
+      if ($hdr !~ /^[^:\s]+:?$/) {
+        $self->lint_warn("config: invalid head test $name header: $hdr");
+        return;
+      }
+      $hdr =~ s/:$//;
+      $conf->{test_opt_header}->{$name} = $hdr;
+      $conf->{test_opt_exists}->{$name} = 1;
     } else {
-      my ($pat) = ($text =~ /^\s*\S+\s*(?:\=|\!)\~\s*(\S.*?\S)\s*$/);
-      if ($pat) { $pat =~ s/\s+\[if-unset:\s+(.+)\]\s*$//; }
-      return unless $self->is_delimited_regexp_valid($name, $pat);
+      if ($text !~ /^([^:\s]+(?:\:|(?:\:[a-z]+){1,2})?)\s*([=!]~)\s*(.+)$/) {
+        $self->lint_warn("config: invalid head test $name: $text");
+        return;
+      }
+      my ($hdr, $op, $pat) = ($1, $2, $3);
+      $hdr =~ s/:$//;
+      if ($pat =~ s/\s+\[if-unset:\s+(.+)\]$//) {
+        $conf->{test_opt_unset}->{$name} = $1;
+      }
+      my ($rec, $err) = compile_regexp($pat, 1, $ignore_amre);
+      if (!$rec) {
+        $self->lint_warn("config: invalid regexp for $name '$pat': $err", $name);
+        return;
+      }
+      $conf->{test_qrs}->{$name} = $rec;
+      $conf->{test_opt_header}->{$name} = $hdr;
+      $conf->{test_opt_neg}->{$name} = 1 if $op eq '!~';
     }
   }
   elsif ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS)
   {
-    return unless $self->is_meta_valid($name, $text);
+    if ($self->is_meta_valid($name, $text)) {
+      # Untaint now once and not repeatedly later
+      $text = untaint_var($text);
+    } else {
+      return;
+    }
   }
 
   $conf->{tests}->{$name} = $text;
   $conf->{test_types}->{$name} = $type;
+
+  if ($name =~ /AUTOLEARNTEST/i) {
+     dbg("config: auto-learn: $name has type $type = $conf->{test_types}->{$name} during add_test\n");
+  }
+
+  
   if ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS) {
     $conf->{priority}->{$name} ||= 500;
   }
@@ -1193,30 +1343,41 @@ sub add_regression_test {
 sub is_meta_valid {
   my ($self, $name, $rule) = @_;
 
+  # $meta is a degenerate translation of the rule, replacing all variables (i.e. rule names) with 0. 
   my $meta = '';
-  $rule = untaint_var($rule);  # must be careful below
+
+  # Paranoid check (Bug #7557)
+  if ($rule =~ /(?:\:\:|->)/)  {
+    warn("config: invalid meta $name rule: $rule") ;
+    return 0;
+  }
+
+  # Process expandable functions before lexing
+  $rule =~ s/${META_RULES_MATCHING_RE}/ 0 /g;
 
   # Lex the rule into tokens using a rather simple RE method ...
-  my $lexer = ARITH_EXPRESSION_LEXER;
-  my @tokens = ($rule =~ m/$lexer/g);
-  if (length($name) == 1) {
-    for (@tokens) {
-      print "$name $_\n "  or die "Error writing token: $!";
-    }
-  }
+  my @tokens = ($rule =~ /($ARITH_EXPRESSION_LEXER)/og);
+
   # Go through each token in the meta rule
   foreach my $token (@tokens) {
-    # Numbers can't be rule names
-    if ($token =~ /^(?:\W+|[+-]?\d+(?:\.\d+)?)$/) {
+    # If the token is a syntactically legal rule name, make it zero
+    if ($token =~ IS_RULENAME) {
+      $meta .= "0 ";
+    }
+    # if it is a (decimal) number or a string of 1 or 2 punctuation
+    # characters (i.e. operators) tack it onto the degenerate rule
+    elsif ($token =~ /^(\d+(?:\.\d+)?|[[:punct:]]{1,2})\z/s) {
       $meta .= "$token ";
     }
-    # Zero will probably cause more errors
+    # Skip anything unknown (Bug #7557)
     else {
-      $meta .= "0 ";
+      $self->lint_warn("config: invalid meta $name token: $token", $name);
+      return 0;
     }
   }
 
-  my $evalstr = 'my $x = ' . $meta . '; 1;';
+  $meta = untaint_var($meta); # was carefully checked
+  my $evalstr = 'my $x = '.$meta.'; 1;';
   if (eval $evalstr) {
     return 1;
   }
@@ -1227,94 +1388,21 @@ sub is_meta_valid {
   return 0;
 }
 
+# Deprecated functions, leave just in case..
 sub is_delimited_regexp_valid {
-  my ($self, $name, $re) = @_;
-
-  if (!$re || $re !~ /^\s*m?(\W).*(?:\1|>|}|\)|\])[a-z]*\s*$/) {
-    $re ||= '';
-    $self->lint_warn("config: invalid regexp for rule $name: $re: missing or invalid delimiters\n", $name);
-    return 0;
-  }
-  return $self->is_regexp_valid($name, $re);
+  my ($self, $rule, $re) = @_;
+  warn "deprecated is_delimited_regexp_valid() called, use compile_regexp()\n";
+  my ($rec, $err) = compile_regexp($re, 1, 1);
+  return $rec;
 }
-
 sub is_regexp_valid {
-  my ($self, $name, $re) = @_;
-
-  # OK, try to remove any normal perl-style regexp delimiters at
-  # the start and end, and modifiers at the end if present,
-  # so we can validate those too.
-  my $origre = $re;
-  my $safere = $re;
-  my $mods = '';
-  local ($1,$2);
-  if ($re =~ s/^m{//) {
-    $re =~ s/}([a-z]*)$//; $mods = $1;
-  }
-  elsif ($re =~ s/^m\(//) {
-    $re =~ s/\)([a-z]*)$//; $mods = $1;
-  }
-  elsif ($re =~ s/^m<//) {
-    $re =~ s/>([a-z]*)$//; $mods = $1;
-  }
-  elsif ($re =~ s/^m(\W)//) {
-    $re =~ s/\Q$1\E([a-z]*)$//; $mods = $1;
-  }
-  elsif ($re =~ s/^\/(.*)\/([a-z]*)$/$1/) {
-    $mods = $2;
-  }
-  else {
-    $safere = "m#".$re."#";
-  }
-
-  if ($self->{conf}->{lint_rules} ||
-      $self->{conf}->{ignore_always_matching_regexps})
-  {
-    my $msg = $self->is_always_matching_regexp($name, $re);
-
-    if (defined $msg) {
-      if ($self->{conf}->{lint_rules}) {
-        $self->lint_warn($msg, $name);
-      } else {
-        warn $msg;
-        return 0;
-      }
-    }
-  }
-
-  # now prepend the modifiers, in order to check if they're valid
-  if ($mods) {
-    $re = "(?".$mods.")".$re;
-  }
-
-  # note: this MUST use m/...${re}.../ in some form or another, ie.
-  # interpolation of the $re variable into a code regexp, in order to test the
-  # security of the regexp.  simply using ("" =~ $re) will NOT do that, and
-  # will therefore open a hole!
-  if (eval { ("" =~ m#${re}#); 1; }) {
-    return 1;
-  }
-  my $err = $@ ne '' ? $@ : "errno=$!";  chomp $err;
-  $err =~ s/ at .*? line \d.*$//;
-  $self->lint_warn("config: invalid regexp for rule $name: $origre: $err\n", $name);
-  return 0;
+  my ($self, $rule, $re) = @_;
+  warn "deprecated is_regexp_valid() called, use compile_regexp()\n";
+  my ($rec, $err) = compile_regexp($re, 1, 1);
+  return $rec;
 }
-
-# check the pattern for some basic errors, and warn if found
 sub is_always_matching_regexp {
-  my ($self, $name, $re) = @_;
-
-  if ($re =~ /(?<!\\)\|\|/) {
-    return "config: regexp for rule $name always matches due to '||'";
-  }
-  elsif ($re =~ /^\|/) {
-    return "config: regexp for rule $name always matches due to " .
-      "pattern starting with '|'";
-  }
-  elsif ($re =~ /\|(?<!\\\|)$/) {
-    return "config: regexp for rule $name always matches due to " .
-      "pattern ending with '|'";
-  }
+  warn "deprecated is_always_matching_regexp() called\n";
   return;
 }
 
@@ -1339,6 +1427,7 @@ sub add_to_addrlist_rcvd {
   my ($self, $listname, $addr, $domain) = @_;
   my $conf = $self->{conf};
 
+  $domain = lc $domain;
   $addr = lc $addr;
   if ($conf->{$listname}->{$addr}) {
     push @{$conf->{$listname}->{$addr}{domain}}, $domain;
@@ -1359,7 +1448,7 @@ sub remove_from_addrlist {
   my $conf = $self->{conf};
 
   foreach my $addr (@addrs) {
-    delete($conf->{$singlelist}->{$addr});
+    delete($conf->{$singlelist}->{lc $addr});
   }
 }
 
@@ -1368,7 +1457,7 @@ sub remove_from_addrlist_rcvd {
   my $conf = $self->{conf};
 
   foreach my $addr (@addrs) {
-    delete($conf->{$listname}->{$addr});
+    delete($conf->{$listname}->{lc $addr});
   }
 }
 
